import java.io.*;
import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

public class LangTag {
    private static final Pattern pIrregular = Pattern.compile("^((en-GB-oed)|(i-ami)|(i-bnn)|(i-default)|(i-enochian)|(i-hak)|(i-klingon)|(i-lux)|(i-mingo)|(i-navajo)|(i-pwn)|(i-tao)|(i-tay)|(i-tsu)|(sgn-BE-FR)|(sgn-BE-NL)|(sgn-CH-DE))$".toLowerCase());
    private static final Pattern pRegular = Pattern.compile
	("^((art-lojban)|(cel-gaulish)|(no-bok)|(no-nyn)|(zh-guoyu)|(zh-hakka)|(zh-min)|(zh-min-nan)|(zh-xiang))$");
    private static final Pattern pPrivateuse = Pattern.compile
	("^(x(-(\\d|[a-z]){1,8})+)$");
    private static final Pattern pFutureUse = Pattern.compile
	("^([a-z]){4}$");
    private static final Pattern pRegisteredLanguage = Pattern.compile
	("^([a-z]){5,8}$");
    private static final Pattern pIso639Shortest = Pattern.compile
	("^([a-z]){2,3}$");
    private static final Pattern pExtlang1a = Pattern.compile
	("^([a-z]){3}$");
    private static final Pattern pExtlang1b = Pattern.compile
	("^([a-z]){3}-.*");
    private static final Pattern pExtlang2a = Pattern.compile
	("^([a-z]){3}-([a-z]){3}$");
    private static final Pattern pExtlang2b = Pattern.compile
	("^([a-z]){3}-([a-z]){3}-.*");
    private static final Pattern pExtlang3a = Pattern.compile
	("^([a-z]){3}-([a-z]){3}-([a-z]){3}$");
    private static final Pattern pExtlang3b = Pattern.compile
	("^([a-z]){3}-([a-z]){3}-([a-z]){3}-.*");
    private static final Pattern pScript = Pattern.compile
	("^([a-z]){4}$");
    private static final Pattern pRegionIso31661Code = Pattern.compile
	("^([a-z]){2}$");
    private static final Pattern pRegionUNM49 = Pattern.compile
	("^([\\d]){3}$");
    private static final Pattern pVariant = Pattern.compile
	("^((([\\d]|[a-z]){5,8})|((\\d)([\\d]|[a-z]){3}))$");
    private static final Pattern pExtension = Pattern.compile
	("^(([a-w]|[y-z]|[0-9])(-(([a-z])|(\\d)){2,8}))$");
    public ArrayList<SubTag> stl;
    public String input;
    public boolean valid;
    public boolean wellformed;
    public LangTag (String input, Boolean validate) {
	this.input = input.toLowerCase();
	this.stl = new ArrayList<SubTag>(100);
	this.wellformed = true;
	this.valid = true;
	abnfCheck();
	if (validate == true) validate();
    }
    private static void validate() {
	System.out.println("VALIDATE!");
    }
    public SubTag getSubtag(int pos) {
	return stl.get(pos);
    }
    public int getSubtagNum() {
	return stl.size();
 }
    private String getSubtag(String restTag){
	if(restTag.matches(".*-.*")){
	    return restTag.split("-")[0];
	}
	else return restTag;
    }
    private void abnfCheck(){ 
	if(pIrregular.matcher(input).matches()){
	    stl.add(new SubTag(input,"irregular"));
	}  else if (pRegular.matcher(input).matches()){
	    stl.add(new SubTag(input,"regular"));
	} else if (pPrivateuse.matcher(input).matches()){
	    stl.add(new SubTag(input,"privateUse"));
	}
	else languageTest(input);
    }

    private void languageTest (String restTag){
	String st = getSubtag(restTag);
	String newRestTag = restTag.replaceFirst(st +"-","");
	if(pRegisteredLanguage.matcher(st).matches()){
	    stl.add(new SubTag(st,"registeredLanguage"));
	    if(st.concat("-").compareTo(restTag) == 0){
		this.wellformed = false;
		stl.add(new SubTag("-","notWellformed-potentialSubtags:script-region-variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && newRestTag.length()>0) { 
		scriptTest(newRestTag);
	    }
	} else if(pFutureUse.matcher(st).matches()) {
	    stl.add(new SubTag(st,"futureUse"));
	    if(st.concat("-").compareTo(restTag) == 0){
		stl.add(new SubTag("-","notWellformed-potentialSubtags:script-region-variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && newRestTag.length()>0) { 
		scriptTest(newRestTag);
	    }
	} else if(pIso639Shortest.matcher(st).matches()) {
	    stl.add(new SubTag(st,"Iso639Shortest"));
	    if(st.concat("-").compareTo(restTag) == 0){
		stl.add(new SubTag("-","notWellformed-potentialSubtags:extlang-script-region-variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && newRestTag.length()>0) { 
		extlangTest(newRestTag);
	    }
	} else {
	    this.wellformed = false;
	    stl.add(new SubTag(input,"notWellformed-potentialSubtags:primaryLangTag-privateUse-regular-irregular"));
	}
    }
    private void extlangTest (String restTag) { 
	if(pExtlang3a.matcher(restTag).matches() || pExtlang3b.matcher(restTag).matches()){
	    stl.add(new SubTag(restTag.substring(0,11),"extlang3"));	    
	} else if (pExtlang2a.matcher(restTag).matches() || pExtlang2b.matcher(restTag).matches()){
	    stl.add(new SubTag(restTag.substring(0,7),"extlang2"));
	} else if (pExtlang1a.matcher(restTag).matches() || pExtlang1b.matcher(restTag).matches()){
	    stl.add(new SubTag(restTag.substring(0,3),"extlang1"));
	}
	String st = stl.get(stl.size()-1).subtag;
	if(stl.get(stl.size()-1).type.matches("^extlang.*")){
	    if(st.concat("-").compareTo(restTag) == 0){
		this.wellformed = false;
		stl.add(new SubTag("-","notWellformed-potentialSubtags:script-region-variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && restTag.replaceFirst(st + "-","").length()>0) { 
		scriptTest(restTag.replaceFirst(st + "-",""));	     
	    }
	}
	else     scriptTest(restTag);
    }
    private void scriptTest (String restTag) { 
	String st = getSubtag(restTag);
	String newRestTag = restTag.replaceFirst(st +"-","");
	if(pScript.matcher(st).matches()){
	    stl.add(new SubTag(st,"script"));
	    if(st.concat("-").compareTo(restTag) == 0){
		this.wellformed = false;
		stl.add(new SubTag("-","notWellformed-potentialSubtags:region-variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && newRestTag.length()>0) { 
		regionTest(newRestTag, "");
	    } 
	}
	else regionTest(restTag,"extlang-script");
    }
    private void regionTest(String restTag, String potentialSubtags) {
	String st = getSubtag(restTag);
	String newRestTag = restTag.replaceFirst(st +"-","");
	if(pRegionIso31661Code.matcher(st).matches()){
	    stl.add(new SubTag(st,"regionIso31661"));
	    if(st.concat("-").compareTo(restTag) == 0){
		this.wellformed = false;
		stl.add(new SubTag("-","notWellformed-potentialSubtags:variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && newRestTag.length()>0) { 
		variantTest(restTag.replaceFirst(st + "-",""),"");
	    }
	}
	else if (pRegionUNM49.matcher(st).matches()) {
	    stl.add(new SubTag(st,"regionUNM49"));
	    if(st.concat("-").compareTo(restTag) == 0){
		this.wellformed = false;
		stl.add(new SubTag("-","notWellformed-potentialSubtags:variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && newRestTag.length()>0) { 
		variantTest(restTag.replaceFirst(st + "-",""),"");
	    }
	} else variantTest(restTag,potentialSubtags + "-region");	
    }
    private void variantTest(String restTag, String potentialSubtags) {
	String st = getSubtag(restTag);
	String newRestTag = restTag.replaceFirst(st +"-","");
	if(pVariant.matcher(st).matches()){
	    stl.add(new SubTag(st,"variant"));
	    if(st.concat("-").compareTo(restTag) == 0){
		this.wellformed = false;
		stl.add(new SubTag("-","notWellformed-potentialSubtags:variant-extension-privateUse"));
	    } else if (st.compareTo(restTag) != 0 && restTag.replaceFirst(st + "-","").length()>0) { 
		variantTest(restTag.replaceFirst(st + "-",""),"");
	    }
	} else extensionTest(restTag,potentialSubtags + "-variant");
    }
    private void extensionTest(String restTag, String potentialSubtags) {
	String xyz1 = restTag.replaceFirst("(([a-w]|[y-z]|[0-9])(-(([a-z])|(\\d)){2,8}))","");
	String st = restTag.substring(0,restTag.length()-xyz1.length());
	String newRestTag = restTag.replaceFirst(st,"");
	if(pExtension.matcher(st).matches()) {
	    stl.add(new SubTag(st,"extension"));
	    if(newRestTag.startsWith("-") & newRestTag.length()>1){
		extensionTest(newRestTag.substring(1),"");
	    } else if (newRestTag.length()>0){
		this.wellformed = false;
		stl.add(new SubTag(newRestTag,"notWellformed-potentialSubtags:extension-privateUse"));
	    }
	} else if(pPrivateuse.matcher(restTag).matches())
	    {
		stl.add(new SubTag(restTag,"privateUse"));
	    } else{
	    this.wellformed = false;
	    stl.add(new SubTag(restTag,"notWellformed-potentialSubtags:" + potentialSubtags + "-extension-privateUse"));
	}
    }
}